//Title: BeTVOut Class
//Platform: BeOS 5
//Version: 2002.02.19
//Description:
//	A class for controlling BeTVOut-supported hardware.
//

//  Copyright (C) 2002 Kevin H. Patterson

#include "BeTVOut.h"

#include "GraphicsDevice.h"

#include "Log.h"
#include <iostream.h>

BeTVOut::BeTVOut(GraphicsDevice * const uGraphicsDevice)
:
	iGraphicsDevice(uGraphicsDevice),
	
	iInitCheck(false),

	iTVOutDeviceMutex(new BLocker),
		
	iTVOut(false),
	
	iFormat(NTSC_M),
	iRefreshRate(59.94),
	iWidth(640),
	iHeight(480),
	iColorSpace(B_RGB32),
	iOutputAutoDetect(true),
	iOutput(OUTPUT_MUX_ALL),
	iFiltering(false),
	iColorBars(false),
	iOverscan(true),
	iDisplay(DISPLAY_MUX_ALL),
	
	ihTVOutDevice(-1),
	
	iTVOutDeviceType(TV_OUT_DEVICE_UNKNOWN),
	
	iTVOutDeviceData(new riva_data_t),
	iBTType(0xFF),
	iBTVersion(0xFF),
	iCHBus(0),
	iCHAddr(0),
	iCHVersion(0),

	iOutputDetection(0)

{
	Log("BeTVOut::BeTVOut()\n");
	
	if (iGraphicsDevice == NULL) {
		Log("BeTVOut::BeTVOut ERROR: Constructor called with NULL GraphicsDevice.\n");
		return;
	}
	
	if (iTVOutDeviceMutex == NULL) {
		Log("BeTVOut::BeTVOut ERROR: BLocker iTVOutDeviceMutex could not be constructed.\n");
		return;
	}
	
	BAutolock tAutoLocker( iTVOutDeviceMutex );

	int bus0wired = 0, bus1wired = 0,				//no wired busses found
		bus0sane = 0, bus1sane = 0,						//no sane buses
		stat = B_OK;

	ihTVOutDevice = open( "/dev/video/riva_iic/1", O_RDWR );
	//check if driver could be reached...
	if (ihTVOutDevice < 0) {
		Log("BeTVOut::BeTVOut ERROR: riva_iic driver not found.\n" );
		return;
	}
	
	//setup 'magic value' to unlock driver: 
	//(to keep compatibility & prevent crashes with older versions of driver and app...)
	iTVOutDeviceData->drv_version = 0xFF;	
	//notify driver of our iBTVersion:
	//(to prevent old apps from accessing new drivers...)
	iTVOutDeviceData->app_version = static_cast<uint16>(iBTVersion);
	//check iBTVersion of driver...
	stat = ioctl( ihTVOutDevice, rivaVersion, iTVOutDeviceData );
	if (iTVOutDeviceData->drv_version == 0xFF) {
		Log("BeTVOut::BeTVOut riva_iic driver version 0.00\n");
		Log("BeTVOut::BeTVOut ERROR: Incompatible driver version (0.22 or higher needed.)\n" );
		goto exit_close;
	}
	//this works without driver supporting riva_data_t struct:
	//well, it did before I changed the first var-type.. :-(
	Log("BeTVOut::BeTVOut riva_iic driver version %.2f\n", iTVOutDeviceData->drv_version / 100.0 ); 
	if (iTVOutDeviceData->drv_version < 22) {
		Log("BeTVOut::BeTVOut ERROR: Incompatible driver version. (0.22 or higher needed.)\n" );
		goto exit_close;
	}
	//(now we know the driver supports the riva_data_t struct...)
	if (stat) { //if ioctl returned 'NACK'..
		Log("BeTVOut::BeTVOut ERROR: Application too old for use with this driver. "
				"(%.3f or higher needed.)\n", iTVOutDeviceData->app_version / 1000.0 );
		goto exit_close;
	}

	//unlock RIVA extended VGA registers...
	ioctl(ihTVOutDevice, rivaUnlock);

	//check for one or two available (wired) IIC busses.
	//On RIVA TNT and later, two could be there...
	{
		int RIVAexists(0);
		Log("BeTVOut::BeTVOut RIVA IIC bus 0: ");
		iTVOutDeviceData->iic.busnr = 0;
		if (!ioctl( ihTVOutDevice, rivaChkBus, iTVOutDeviceData )) {
			Log("detected.\n");
			RIVAexists = 1;
			bus0wired = 1;
		}	else {
			Log("not detected.\n");
		}
		Log("BeTVOut::BeTVOut RIVA IIC bus 1: ");
		iTVOutDeviceData->iic.busnr = 1;
		if (!ioctl( ihTVOutDevice, rivaChkBus, iTVOutDeviceData )) {
			Log("detected.\n");
			RIVAexists = 1;
			bus1wired = 1;
		} else {
			Log("not detected.\n");
		}
		if (!RIVAexists) {
			Log("BeTVOut::BeTVOut ERROR: No RIVA IIC bus found.\n");
			goto exit_close;
		}
	}

	//find at least one sane IIC bus... (test this by sending an unused adress...)
	if (!FindSaneBus(
			ihTVOutDevice, bus0wired, bus1wired, &bus0sane, &bus1sane, iTVOutDeviceData )) {
		Log("BeTVOut::BeTVOut ERROR: No *sane* RIVA IIC busses detected.\n");
		goto exit_close;
	}

	//now find a TV-out chip...

	//check for BT or CX chip:
	if (FindBT( ihTVOutDevice, iTVOutDeviceData ))
		iTVOutDeviceType = BT;
		
	if (iTVOutDeviceType == BT)
	{
		//read chiptype and iBTVersion number:
		switch (ReadType( ihTVOutDevice, &iBTVersion, &iBTType, iTVOutDeviceData )) {
		case 0:
			//no errors found:
			Log("BeTVOut::BeTVOut ChipID = %d, chip version = %d: Chiptype is ", iBTType, iBTVersion);
			switch (iBTType) {
			case 0:
				Log("bt868.\n");
				break;
			case 1:
				Log("bt869.\n");
				break;
			case 2:
				Log("cx25870.\n");
				break;
			case 3:
				Log("cx25871.\n");
				break;
			default:
				Log("unknown.\nBeTVOut::BeTVOut ERROR: Unknown TV-Out device.\n");
				goto exit_close;
				break;
			}
			break;
		case 1:
			//IIC error encountered:
			goto exit_close;
			break;
		default:
			//The chip apparantly is not a TVout chip:
			iTVOutDeviceType = TV_OUT_DEVICE_UNKNOWN;
			break;
		} 
	}		

	//check for CH chip:
	if (iTVOutDeviceType == TV_OUT_DEVICE_UNKNOWN)
	{
		if (bus0sane) {
			iTVOutDeviceData->iic.busnr = 0;
			iCHBus = 0;
			if (ch_detect( ihTVOutDevice, iCHBus, &iCHAddr, &iCHVersion, iTVOutDeviceData ))
				iTVOutDeviceType = CH;
		}
		if (bus1sane && (iTVOutDeviceType == TV_OUT_DEVICE_UNKNOWN)) {
			iTVOutDeviceData->iic.busnr = 1;
			iCHBus = 1;
			if (ch_detect( ihTVOutDevice, iCHBus, &iCHAddr, &iCHVersion, iTVOutDeviceData ))
				iTVOutDeviceType = CH;
		}
		if (iTVOutDeviceType == CH) {
			Log("BeTVOut::BeTVOut CH TV-Out device detected on bus %d, adress $%x;\n", iCHBus, iCHAddr );
			Log("BeTVOut::BeTVOut Chiptype is ch%s.\n", ch_model_name(iCHVersion) );
		}
		else {	//no TVout chip at all...
			Log("BeTVOut::BeTVOut No supported TVout chip found.\n" );
			goto exit_close;
		}
	}

	iTVOutDeviceData->devID = 0;
	//get current RIVA display mode:
	ioctl( ihTVOutDevice, rivaGetMode, iTVOutDeviceData );

	Log("BeTVOut::BeTVOut RIVA h_display: %d, ", iTVOutDeviceData->displ_mode.displ_timing.h_display );
	Log("v_display: %d\n", iTVOutDeviceData->displ_mode.displ_timing.v_display );

	//RIVA programming should be universal for every chipversion, but
	//the 'horizontal sync phase' setting in the TV-out chip SHOULD be done 
	//seperately for different RIVA chips and videomodes though...
	//(for both CH and BT chips).
	//WARNING: This applies for both VESA and 'driver' (BeScreen) mode!
	Log("BeTVOut::BeTVOut Vendor ID = $%x, Device ID = $%x\n",
		iTVOutDeviceData->dev.vendor_id, iTVOutDeviceData->dev.device_id );
	Log("BeTVOut::BeTVOut nVidia Corporation ");
	switch (iTVOutDeviceData->dev.device_id)
	{
		case 0x0020:
			Log("RIVA TNT1 detected.\n" );
			iTVOutDeviceData->type = 1;
			break;
		case 0x0028:
			Log("RIVA TNT2 detected.\n" );
			iTVOutDeviceData->type = 2;
			break;
		case 0x002D:
			Log("RIVA TNT2-M64 detected.\n" );
			iTVOutDeviceData->type = 3;
			break;
		case 0x0110:
			Log("RIVA GeForce2 MX(200/400) detected.\n" );
			iTVOutDeviceData->type = 1;
			break;
		default:
			Log("unknown card.\n");
			Log("BeTVOut::BeTVOut WARNING: Unknown nVidia card detected.\n");
	}

	iInitCheck = true;
	iTVOutDeviceMutex->Unlock();
	return;
	
exit_close:
	close( ihTVOutDevice );
	ihTVOutDevice = -1;
	return;
}


BeTVOut::~BeTVOut() {
	Log("BeTVOut::~BeTVOut()\n");
	if (iTVOutDeviceMutex != NULL)
		iTVOutDeviceMutex->Lock();

	if (iGraphicsDevice != NULL)
		iGraphicsDevice->BufferCyclerStop();
	
	if (ihTVOutDevice >= 0)
		close( ihTVOutDevice );
	ihTVOutDevice = -1;
	
	delete iTVOutDeviceData;
	iTVOutDeviceData = NULL;
	
	delete iTVOutDeviceMutex;
}


bool BeTVOut::InitCheck() const {
	return iInitCheck;
}


bool BeTVOut::Lock() {
	return iTVOutDeviceMutex->Lock();
}


status_t BeTVOut::LockWithTimeout( bigtime_t uTimeout ) {
	return iTVOutDeviceMutex->LockWithTimeout( uTimeout );
}


void BeTVOut::Unlock() {
	iTVOutDeviceMutex->Unlock();
}


status_t BeTVOut::TVOut( const bool uEnable ) {
	Log("BeTVOut::TVOut()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );

	//TV-Out won't be enabled until the end of this function.
	iTVOut = false;

	//BETA: Various filtering levels need to be supported (not just ON/OFF).
	const unsigned char tFiltering( (iFiltering != 0) & 0x01 );

	//A variable to hold a unique 'BeTVOut Mode'. (consistent with BeTVOut mode menu)
	const unsigned char tMode( pBeTVOutMode() );
	if (tMode < 1) {
		Log("BeTVOut::TVOut ERROR: Unsupported BeTVOut Mode.\n");
		goto exit_error;
	}	

	display_mode new_mode;
	iGraphicsDevice->GetMode( &new_mode );	//'save' current mode...

	//Reset the TV-Out chip...
	switch (iTVOutDeviceType) {
	case BT:
		//BETA: reset BT should be here...
		//(but: beware of the 'locked SDA' syndrome then!)
		if (KillClock_Blackout_BT( ihTVOutDevice, iBTType, iTVOutDeviceData ))
			goto exit_error;
		break;
	case CH:
		//BETA: We don't know how to reset the CH chips?
		//reset CH or use another way to make sure all registers get set the same
		//way non-dependent of how they were set before BeTVOut was started...
		break;
	default:
		//We don't know how to reset other types of TV-Out chips...
		break;
	}

	//Perform Auto Detection on video Outputs...
	pOutputAutoDetect();

	//Program the TV-Out chip...
	if (iColorBars) {

		//This is all the programming that is necessary for color bars!
		switch (iTVOutDeviceType) {
		case BT:
			switch (tMode) {
			case 1:
				if (InitBT_NTSC640( ihTVOutDevice, iTVOutDeviceData ))
					goto exit_error;
				break;
			case 4:
				if (InitBT_PAL800( ihTVOutDevice, iTVOutDeviceData ))
					goto exit_error;
				break;
			default:
				goto exit_error;
				break;
			}
			break;
		case CH:
			goto exit_error;
			break;
		default:
			goto exit_error;
			break;
		}
		
	} else {

		//A 'REAL' TV-Out Mode has been selected, so do the programming...
		
		//First, 'slowdown RIVA clock'...
		//makes sure TVout chip is not being overclocked! (VESA640/800 is safe..)
		//NOTE: it looks like this is unnesessary after all..
		if ((iWidth * iHeight) <= 307200) 
			nonVESAsetVESA640(new_mode);
		else
			nonVESAsetVESA800(new_mode);
		
		switch (iTVOutDeviceType) {
		case BT:
			
			{
				//Manual config is non-dependent of the state of the PAL hardware input pin;
				//Also SDA lockups occur when setting EN_XCLK after autoconfig!
				//Make sure PAL_MD=0 for NTSC and PAL_MD = 1 for PAL...
				switch (tMode) {
				case 2:
					if (InitBT_NTSC640( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 3:
					if (InitBT_NTSC800( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 5:
					if (InitBT_PAL640( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 6:
					if (InitBT_PAL800( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 9:
					if (InitBT_NTSC720( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 10:
					if (InitBT_PAL720( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 11:
					if (InitBT_NTSC640_OS( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				case 12:
					if (InitBT_PAL800_OS( ihTVOutDevice, iTVOutDeviceData ))
						goto exit_error;
					break;
				}
				//modify BT Hphase signal to center TV image...
				if (SetBT_Hphase( ihTVOutDevice, tMode, iTVOutDeviceData ))
					goto exit_error;
			}
			break;
			
		case CH:

			{
				//For TNT2 chips horizontal position should be shifted right by 8 pixels.
				//(all values tested on TNT2 and GeForce2 cards.)
				unsigned char hoffset = 2;		//2 x 4 pixels: TNT2 or TNT2 M64 
				if (iTVOutDeviceData->type == 1) hoffset = 0;	//TNT1 or GeForce2
				
				//Because DVD modes with Chrontel are critical a special correction
				//needs to be done for these modes if RIVA is a TNT1 or GeForce2...
				//(It looks like the non-M64/M64 patch is needed, but this doesn't seem to
				// work here. The non-M64/M64 patch correction is normally visible only on
				// non-overscanning modes, so it only needs to be done there if these
				// modes will be set using BeScreen which currently isn't the case.
				// The currently used direct register programming method *is* using the
				// non-M64/M64 patch.
				// NOTE that the normally used 'hoffset' settings are set for centered TV
				// output, including for the Chrontel DVD modes. Only there we get
				// viewport problems if we don't apply the patch below:)
				if ( iOverscan &
						( ((iWidth == 720) && (iHeight = 480))
						| ((iWidth == 720) && (iHeight = 576)) )
						& (iTVOutDeviceData->type == 1) )
					hoffset = 2;
				
				switch (tMode) {
				case 2:
					//NTSC640 is mode 17, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,17,(0x30 + hoffset),
						0x02,0x01, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 3:
					//NTSC800 resembles mode 24.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,24,(0x3e + hoffset),
						0xff,0x00, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 5:
					//PAL640 resembles mode 19.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,19,(0x38 + hoffset),
						0x00,0x00, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 6:
					//PAL800 is mode 20, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,20,(0x40 + hoffset),
						0x00,0x00, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 9:
					//CH DVD MODES NOTICE:
					//The official CH DVD modes are interlaced input only, while
					//the RIVA cannot output interlaced data. This is why these
					//CH modes (25 and 26) cannot be used..  :-(
					//SO: if you want 'perfect' TV output in non-standard modes:
					//get a card with Brooktree BT868/869 chip mounted on it...
					
					//NTSC720 'resembles' mode 16.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: do not change Hoffset: CH viewport is critical,
					//because we 'display' 720 pixels on a 640 pixel wide viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,16,(0x16 + hoffset),
						0x00,0x01, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 10:
					//PAL720 resembles mode 19.
					//(tested on TNT2-M64 with CH7006 only: has to be tested more..)
					//WARNING: be carefull if Hoffset in changed: CH H_overscan is
					//critical, because we display 720 pixels in a 800 pixel
					//viewport.
					//(current values are (and have to be) exactly 'balanced'.)
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,19,(0x14 + hoffset),
						0x00,0x00, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 11:
					//NTSC640_OS is mode 16, this is compatible to BT and RIVA timing.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,16,(0x22 + hoffset),
						0x00,0x01, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				case 12:
					//PAL768 resembles mode 19.
					ch_rud_set_mode(
						ihTVOutDevice, iCHBus, iCHAddr,19,(0x1a + hoffset),
						0x00,0x00, tFiltering, iTVOutDeviceData);
						//mode, horpos, verpos, oflow
					break;
				}
				//BTW: I checked the videosignals on a scope:
				//all used modes are all perfectly OK!
		
				//BTW 2: The CH700X seem to be much more forgiving than the BT chips!
				//If you change just a little in the wrong way on the BT your TV and VGA
				//image dissappear, while the CH does not suffer from these things!
				//Also the CH pixelclock may be as high as 50Mhz, while BT max is only 40Mhz..
				//The downside: The CH only supports the predefined modes, while the BT
				//is (almost) fully programmable...
		
				//BTW3: Above picture modes and position are tested on a
				//ASUS AGP 32Mb V7100 GeForce2 MX200 with CH7007 TVout chip,
				//and on a TNT2 card.
		
				//modify Hphase signal to center TV image...
				//already done in 'ch_rud_set_mode' above.
		
				//setup CH for TVout (filters...)
			}
			break;
			
		default:
			
			goto exit_error;
			break;
			
		}

	}

	//Program the RIVA...
	switch (tMode) {
	case 1:
	case 4:
		if (DoTestImage( ihTVOutDevice, iTVOutDeviceData ))
			goto exit_error;
   	Log(" and NTSC testimage activated.\n");
		break;
	case 2:
		//reprogram RIVA for NTSC 640x480 @ 60Hz mode
		ioctl(ihTVOutDevice, rivaNtsc640, iTVOutDeviceData);
		Log(" and NTSC 640x480 @ 60Hz mode activated.\n");
		break;
	case 3:
		//reprogram RIVA for NTSC 800x600 @ 60Hz mode
		ioctl(ihTVOutDevice, rivaNtsc800, iTVOutDeviceData);
		Log(" and NTSC 800x600 @ 60Hz mode activated.\n");
		break;
	case 5:
		//reprogram RIVA for PAL 640x480 @ 50Hz mode
		ioctl(ihTVOutDevice, rivaPal640, iTVOutDeviceData);
		Log(" and PAL 640x480 @ 50Hz mode activated.\n");
		break;
	case 6:
		//reprogram RIVA for PAL 800x600 @ 50Hz mode
		ioctl(ihTVOutDevice, rivaPal800, iTVOutDeviceData);
		Log(" and PAL 800x600 @ 50Hz mode activated.\n");
		break;
	case 9:
		if (iTVOutDeviceType == BT) {
			//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
			//(tested on TNT2 with BT869)
			new_mode.virtual_width = 720;			//H_ACTIVE
			new_mode.timing.h_display = 720;
			new_mode.virtual_height = 480;			//V_ACTIVEI
			new_mode.timing.v_display = 480;

			new_mode.timing.h_sync_start = 744;		//do not change!
			new_mode.timing.h_sync_end = 744+144;	//delta is H_sync_pulse
			new_mode.timing.h_total = 888;			//BT H_TOTAL
			new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
			new_mode.timing.v_sync_end = 490+26;	//delta is V_sync_pulse
			new_mode.timing.v_total = 525;			//CH V_TOTAL
			new_mode.timing.pixel_clock = uint32((double(888 * 525) * iRefreshRate) / 1000.0); //refresh
		}	else { //CH found
			//reprogram RIVA for NTSC 720x480 @ 60Hz mode:
			//(tested on TNT2-M64 with CH7006)
			new_mode.virtual_width = 720;			//H_ACTIVE
			new_mode.timing.h_display = 720;
			new_mode.virtual_height = 480;			//V_ACTIVEI
			new_mode.timing.v_display = 480;

			new_mode.timing.h_sync_start = 744;		//do not change! (CH viewport is critical...)
			new_mode.timing.h_sync_end = 744+20;	//delta is H_sync_pulse
			new_mode.timing.h_total = 784;			//CH H_TOTAL
			new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
			new_mode.timing.v_sync_end = 490+25;	//delta is V_sync_pulse
			new_mode.timing.v_total = 525;			//CH V_TOTAL
			new_mode.timing.pixel_clock = uint32((double(784 * 525) * iRefreshRate) / 1000.0); //refresh
		}
		iGraphicsDevice->SetMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(ihTVOutDevice, rivaEnableTVOut);
		Log(" and NTSC 720x480 @ 60Hz overscanning mode activated.\n");
		break;
	case 10:
		//reprogram RIVA for PAL 720x576 @ 50Hz mode:(a bit modified for CH)
		new_mode.virtual_width = 720;			//BT H_ACTIVE
		new_mode.timing.h_display = 720;
		new_mode.virtual_height = 576;			//BT  V_ACTIVEI
		new_mode.timing.v_display = 576;
		new_mode.timing.h_sync_start = 744;		//set for centered sync pulse
		new_mode.timing.h_sync_end = 744+140;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 888;			//BT H_CLKI
		new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 625;			//BT V_LINESI
		new_mode.timing.pixel_clock = uint32((double(888 * 625) * iRefreshRate) / 1000.0); //refresh
		iGraphicsDevice->SetMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(ihTVOutDevice, rivaEnableTVOut);
		Log(" and PAL 720x576 @ 50Hz overscanning mode activated.\n");
		break;
	case 11:
		//reprogram RIVA for NTSC 640x480 (OverScanning) @ 60Hz mode:
		new_mode.virtual_width = 640;			//BT H_ACTIVE
		new_mode.timing.h_display = 640;
		new_mode.virtual_height = 480;			//BT  V_ACTIVEI
		new_mode.timing.v_display = 480;
		new_mode.timing.h_sync_start = 744;		//set for CH/BT compatible TV output
		new_mode.timing.h_sync_end = 744+20;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 784;			//BT H_CLKI
		new_mode.timing.v_sync_start = 490;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 490+25;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 525;			//BT V_LINESI
		new_mode.timing.pixel_clock = uint32((double(784 * 525) * iRefreshRate) / 1000.0); //refresh
		iGraphicsDevice->SetMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(ihTVOutDevice, rivaEnableTVOut);
		Log(" and NTSC 640x480 @ 60Hz overscanning mode activated.\n");
		break;
	case 12:
		//reprogram RIVA for PAL 768x576 @ 50Hz mode:
		new_mode.virtual_width = 768;			//H_ACTIVE
		new_mode.timing.h_display = 768;
		new_mode.virtual_height = 576;			//V_ACTIVEI
		new_mode.timing.v_display = 576;
		new_mode.timing.h_sync_start = 848;		//set for centered TV output
		new_mode.timing.h_sync_end = 848+20;	//delta is BT H_BLANKI
		new_mode.timing.h_total = 944;			//BT H_CLKI
		new_mode.timing.v_sync_start = 579;		//set for centered sync pulse
		new_mode.timing.v_sync_end = 579+42;	//delta is BT V_BLANKI
		new_mode.timing.v_total = 625;			//BT V_LINESI
		new_mode.timing.pixel_clock = uint32((double(944 * 625) * iRefreshRate) / 1000.0); //refresh
		iGraphicsDevice->SetMode( &new_mode );
		//switch RIVA to slave-mode for TVout chip:
		ioctl(ihTVOutDevice, rivaEnableTVOut);
		Log(" and PAL 768x576 @ 50Hz overscanning mode activated.\n");
		break;
	}

	pOutput();
	pDisplay();

	iTVOut = uEnable;

	iTVOutDeviceMutex->Unlock();
	return B_OK;

exit_error:
	return B_ERROR;
}


bool BeTVOut::TVOut() const {
	return iTVOut;
}


status_t BeTVOut::Format( const video_format uFormat ) {
	Log("BeTVOut::Format()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	//Check the requested Video Format...
	switch (uFormat) {
	case NTSC_M:
		iFormat = uFormat;
		iRefreshRate = 59.94;
		return B_OK;
		break;
	case NTSC_J:
		iFormat = uFormat;
		iRefreshRate = 60.00;
		return B_OK;
		break;
	case PAL_BDGHI:
	case PAL_M:
	case PAL_N:
	case SECAM:
		iFormat = uFormat;
		iRefreshRate = 50.00;
		return B_OK;
		break;
	default:
		Log("BeTVOut::Format ERROR: Unknown Video Format.\n");
		return B_ERROR;
		break;
	}
}


BeTVOut::video_format BeTVOut::pSimplifiedFormat() const {
	//Check the current Video Format and determine the 'simplified' Video Format...
	//This function returns either NTSC, PAL, or VIDEO_FORMAT_UNKNOWN.
	switch (iFormat) {
	case NTSC_M:
	case NTSC_J:
		return NTSC;
		break;
	case PAL_BDGHI:
	case PAL_M:
	case PAL_N:
	case SECAM:
		return PAL;
		break;
	default:
		return VIDEO_FORMAT_UNKNOWN;
		break;
	}
}


BeTVOut::video_format BeTVOut::Format() const {
	return iFormat;
}


status_t BeTVOut::Resolution( const int32 uWidth, const int32 uHeight ) {
	Log("BeTVOut::Resolution()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iWidth = uWidth;
	iHeight = uHeight;
	return B_OK;
}


status_t BeTVOut::ColorSpace( const color_space uColorSpace) {
	Log("BeTVOut::ColorSpace()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	switch (uColorSpace) {
	case B_GRAY8:
		iColorSpace = B_CMAP8;
		return B_OK;
		break;
	case B_CMAP8:
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
	case B_RGB24:
	case B_RGB32:
	case B_RGBA32:
		iColorSpace = uColorSpace;
		return B_OK;
		break;
	default:
		return B_ERROR;
		break;
	}
	return B_ERROR;
}


color_space BeTVOut::ColorSpace() const {
	return iColorSpace;
}


status_t BeTVOut::ProposeMode(
	video_format &uFormat,
	float &uFrameRate,
	int32 &uWidth,
	int32 &uHeight,
	color_space &uColorSpace ) const
{
	Log("BeTVOut::ProposeMode()\n");
	switch (uColorSpace) {
	case B_GRAY8:
	case B_CMAP8:
	case B_RGB15:
	case B_RGBA15:
	case B_RGB16:
	case B_RGB24:
	case B_RGB32:
	case B_RGBA32:
		break;
	default:
		uColorSpace = B_RGB32;
		break;
	}

	if ((uFormat < NTSC_M) || (uFormat > SECAM))
		uFormat = VIDEO_FORMAT_UNKNOWN;
	
	if (uFormat == VIDEO_FORMAT_UNKNOWN) {
		if (uFrameRate == 0.0) {
			if (uHeight == 0) {
				uFormat = NTSC;
				uFrameRate = 59.94;
				uWidth = 640;
				uHeight = 480;
				return B_OK;
			} else {
				if (uHeight > 525) {
					uFormat = PAL;
					uFrameRate = 50.0;
				} else {
					uFormat = NTSC;
					uFrameRate = 59.94;
				}
			}
		}
	} else {
		switch (uFormat) {
		case NTSC_M:
			uFrameRate = 59.94;
			break;
		case NTSC_J:
			uFrameRate = 60.0;
			break;
		default:
			uFrameRate = 50.0;
		}
	}
	
	if (uHeight < 528) {
		uHeight = 480;
		if (uWidth > 680) {
			uWidth = 720;
			return B_OK;
		} else {
			uWidth = 640;
			return B_OK;
		}
	}

	if (uHeight < 588) {
		uHeight = 576;
		if (uWidth > 744) {
			uWidth = 768;
			return B_OK;
		} else {
			uWidth = 720;
			return B_OK;
		}
	}

	uHeight = 600;
	uWidth = 800;
	return B_OK;
}


status_t BeTVOut::OutputAutoDetect( const bool uOutputAutoDetect ) {
	Log("BeTVOut::OutputAutoDetect()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iOutputAutoDetect = uOutputAutoDetect;
	if (iTVOut)
		if (pOutputAutoDetect() >= B_OK) {
			return pOutput();
		}
	return B_ERROR;
}


status_t BeTVOut::pOutputAutoDetect() {
	if (iOutputAutoDetect) {
		switch (iTVOutDeviceType) {
		case BT:
			//if TV output signals is set to 'autodetect' then do it:
			iOutputDetection = ReadBT_ConnectStat(ihTVOutDevice, iTVOutDeviceData);
			if (iColorBars) {
				if (DoTestImage( ihTVOutDevice, iTVOutDeviceData ))
					return B_ERROR;
			}
			if (iOutputDetection != -1)
				return B_OK;
			break;
		case CH:
			//if TV output signals is set to 'autodetect' then do it:
			//(make sure this does only gets called if CH input clock gets set below
			// its speed limits afterwards, or CH might get destroyed...)
			iOutputDetection = ch_detect_connection(ihTVOutDevice, iCHBus, iCHAddr, iTVOutDeviceData);
			Log("BeTVOut::pOutputAutoDetect TV-Out monitor status = %d\n", iOutputDetection );
			return B_OK;
			//iOutputDetection value is:
			//0: Y/C and CVBS is connected
			//1: only Y/C is connected
			//6: only CVBS is connected
			//7: nothing is connected.
			//(tested with ASUS V7100 GeForce2 MX200 AGP 32Mb.)
			break;
		default:
			return B_ERROR;
			break;
		}
	}
	return B_ERROR;
}


bool BeTVOut::OutputAutoDetect() const {
	return iOutputAutoDetect;
}


status_t BeTVOut::Output( const output_mux uOutput ) {
	Log("BeTVOut::Output()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iOutput = uOutput;
	if (iTVOut)
		return pOutput();
	return B_ERROR;
}


status_t BeTVOut::pOutput() {
	switch (iTVOutDeviceType) {
	case BT:
		//setup BT for TVout (filters, DAC_muxes...)
		if (
			!SetupBTforTVout(
				ihTVOutDevice, static_cast<unsigned char>(iOutput & 0x03), iOutputDetection,
				static_cast<unsigned char>((iFiltering > 0) & 0x01), iTVOutDeviceData )
			)
			return B_OK;
		break;
	default:
		//BETA: Need CH support!
		return B_ERROR;
	}
	return B_ERROR;
}


BeTVOut::output_mux BeTVOut::Output() const {
	return iOutput;
}


status_t BeTVOut::Filtering( const int32 uFiltering ) {
	Log("BeTVOut::Filtering()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iFiltering = uFiltering;
	if (iTVOut) {
		return pOutput();
	} else {
		//TVOut is NOT currently enabled
		//The requested state has been stored for the next TVOut mode...
		return B_OK;
	}
	return B_ERROR;
}


int32 BeTVOut::Filtering() const {
	return iFiltering;
}


status_t BeTVOut::ColorBars( const bool uColorBars) {
	Log("BeTVOut::ColorBars()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iColorBars = uColorBars;
	return B_OK;
}


bool BeTVOut::ColorBars() const {
	return iColorBars;
}


status_t BeTVOut::Overscan( const bool uOverscan ) {
	Log("BeTVOut::Overscan()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iOverscan = uOverscan;
	return B_OK;
}


bool BeTVOut::Overscan() const {
	return iOverscan;
}


status_t BeTVOut::Display( const display_mux uDisplay ) {
	Log("BeTVOut::Display()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	iDisplay = uDisplay;
	if (iTVOut) {
		return pDisplay();
	} else {
		//TVOut is NOT currently enabled
		//The requested state has been stored for the next TVOut mode...
		return B_OK;
	}
	return B_ERROR;
}


status_t BeTVOut::pDisplay() {
	if (iDisplay) {
		//Turn on VGA output:
		ioctl(ihTVOutDevice, rivaVgaOn);
		return B_OK;
	} else {
		//Turn off VGA output:
		ioctl(ihTVOutDevice, rivaVgaOff);
		return B_OK;
	}
}


BeTVOut::display_mux BeTVOut::Display() const {
	return iDisplay;
}


status_t BeTVOut::Field( int32 &uCurrentField ) const {
	//Log("BeTVOut::Field()\n");
	BAutolock tAutoLocker( iTVOutDeviceMutex );
	if (ihTVOutDevice < 0)
		return B_ERROR;
	uCurrentField = ReadBT_Field( ihTVOutDevice, iTVOutDeviceData );
	return B_OK;
}

const unsigned char BeTVOut::pBeTVOutMode() const {
	//Determine which 'BeTVOut Mode' matches the current parameters...
	//Returns a valid BeTVOut Mode number or '0' if a matching mode could not be found.
	int32 tFormat( pSimplifiedFormat() );
	if (iColorBars) {
		switch (tFormat) {
		case NTSC:
			return 1;
			break;
		case PAL:
			return 4;
			break;
		}
	} else {
		switch (tFormat) {
		case NTSC:
			if (iOverscan) {
				if ((iWidth == 640) && (iHeight == 480))
					return 11;
				else {
					if ((iWidth == 720) && (iHeight == 480))
						return 9;
					else {
						Log("BeTVOut::pBeTVOutMode ERROR: Unsupported Resolution for NTSC Overscan.\n");
						return 0;
					}
				}
			} else {
				if ((iWidth == 640) && (iHeight == 480))
					return 2;
				else {
					if ((iWidth == 800) && (iHeight == 600))
						return 3;
					else {
						Log("BeTVOut::pBeTVOutMode ERROR: Unsupported Resolution for NTSC Non-Overscan.\n");
						return 0;
					}
				}
			}
			break;
		case PAL:
			if (iOverscan) {
				if ((iWidth == 720) && (iHeight == 576))
					return 10;
				else {
					if ((iWidth == 768) && (iHeight == 576))
						return 12;
					else {
						Log("BeTVOut::pBeTVOutMode ERROR: Unsupported Resolution for PAL Overscan.\n");
						return 0;
					}
				}
			} else {
				if ((iWidth == 640) && (iHeight == 480))
					return 5;
				else {
					if ((iWidth == 800) && (iHeight == 600))
						return 6;
					else {
						Log("BeTVOut::pBeTVOutMode ERROR: Unsupported Resolution for PAL Non-Overscan.\n");
						return 0;
					}
				}
			}
			break;
		}
	}
	Log("BeTVOut::pBeTVOutMode ERROR: Unsupported Video Format.\n");
	return 0;
}


int BeTVOut::nonVESAsetVESA640( display_mode &new_mode ) {
	//set VESA640: (hex values are exact copies of direct register programmings)
	new_mode.space = iColorSpace;
	new_mode.virtual_width = 640;
	new_mode.virtual_height = 480;
	new_mode.timing.h_display = 640;
	new_mode.timing.v_display = 480;
	
	new_mode.timing.h_sync_start = (0x53 * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1f + (0x53 && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x5F * 8);//byte oriented! (val. *must* be set!)
	new_mode.timing.v_sync_start = 0x1ea;
	new_mode.timing.v_sync_end = ((0x1ea && 0xff0) + 0x00c);
	new_mode.timing.v_total = 0x20b;
	new_mode.timing.pixel_clock = uint32((double((0x5F * 8) * 0x20b) * iRefreshRate) / 1000.0);// in kHz!

	return iGraphicsDevice->SetMode( &new_mode );
}


int BeTVOut::nonVESAsetVESA800( display_mode &new_mode ) {
	//set VESA800: (hex values are exact copies of direct register programmings)
	new_mode.space = iColorSpace;
	new_mode.virtual_width = 800;
	new_mode.virtual_height = 600;
	new_mode.timing.h_display = 800;
	new_mode.timing.v_display = 600;
	
	new_mode.timing.h_sync_start = (0x6a * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1a + (0x6a && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x7f * 8);//byte oriented! (this value *must* be set!)
	new_mode.timing.v_sync_start = 0x259;
	new_mode.timing.v_sync_end = ((0x259 && 0xff0) + 0x00d);
	new_mode.timing.v_total = 0x272;
	new_mode.timing.pixel_clock = uint32((double((0x7F * 8) * 0x272) * iRefreshRate) / 1000.0);// in kHz!

	return iGraphicsDevice->SetMode( &new_mode );
}
